// Listing9-8.S
//
// 32-bit ARM code for the
// decimal string to numeric conversion.
// Ludicrous version (crazy table size).
//
// Written to run on a Raspberry Pi.

#include    "ao32aa.inc"

// Number of iterations for testing:

#define iterations 100000000

// The following are required to use ARM V7&8 instructions
// on Raspberry Pi devices (3, 4, & 5):

            .arch   armv8-a

// Manifest constant for tab character:

#define tab 0x9

        .data
    
// Data to convert:
  
in0:    .asciz  "0"
in1:    .asciz  "1"
in2:    .asciz  "22"
in3:    .asciz  "333"
in4:    .asciz  "4444"
in4a:   .asciz  "1234"
in4b:   .asciz  "5678"
in5:    .asciz  "55555"
in5a:   .asciz  "12345"
in5b:   .asciz  "67890"
in6:    .asciz  "666666"
in7:    .asciz  "7777777"
in8:    .asciz  "88888888"
in9:    .asciz  "999_999_999"
in10:   .asciz  "4_294_967_295"
in11:   .asciz  " ;:,\t 12345678 "
in12:   .asciz  "1_999_999_999"

// Pointers to strings, so asmMain can process
// all of them in a loop:
        
strPtrs:
        .word   in0, in1, in2, in3
        .word   in4, in4a, in4b
        .word   in5, in5a, in5b
        .word   in6, in7
        .word   in8, in9, in10, in11, in12
        .set    evCnt, (.-strPtrs)/4    //Number of strings

eve:    .asciz  ""
  		.align	2  
evf:    .asciz  "*1234"
  		.align	2  
evg:    .asciz  "12*34"
  		.align	2  
evh:    .asciz  "12345678900"


// Pointers to strings, so asmMain can process

        .text
        .align  2

        #include    "print.inc"        

// getTitle returns the address of a string in
// the R0 register. The c.cpp program will print
// this string before and after calling asmMain.

        .text
        .align  2
        
// Here's the title string that c.cpp will print:

title:  wastr  "Listing9-8"
        
        proc    getTitle, public
        adr     r0, title
        bx      lr
        endp    getTitle

////////////////////////////////////////////////////
//
// bufToU32
//
// Decimal string to 32-bit integer conversion.
// This code allows underscores to appear within a
// hex string (but underscores cannot be the first
// character of a decimal string). This code also
// allows an underscore at the end of the string;
// if you consider this to be an error, modify
// the code at label "delimiter" to test for an
// underscore in the previous character position.
//
// On entry:
//
//  R0-     Pointer to buffer holding string.
//
// On exit-
//
//  R0-     Points at the first character beyond
//          the converted string, if conversion
//          was successful.
//
//  R1-     Contains the converted value if the
//          conversion was successful.
//
//  Carry-  Clear if conversion was successful.
//          Set if unsuccessful. If unsuccessful
//          conversion, then R0 will point at
//          the character the conversion failed
//          on and R1 will contain one of the
//          following error codes:
//
//          -1: First (non-WS) character was not
//              a valid decimal digit.
//
//          -2: Overflow (value did not fit into
//              32 bits).
//
//          -3: Last character was not a valid
//              delimiter character (this could
//              be considered a "soft" error).
//
//  This function will skip any leading whitespace
// (spaces or tabs) appearing at the beginning
// of the string.
        
#define delimCode 11
#define underscore 10
#define illChar 0xff

    .macro  entry
    push    {r2-r6, lr}
    .endm
    
    .macro  rtn
    pop     {r2-r6, pc}
    .endm

// Note: NOT thumb code.

    proc    bufToU32, public
    entry
    
    ldr     r3, =429496729 	// For overflow check.
    ldr     r4, =classTbl   // For char classification.
    mov     r5, #100        // For radix-100 multiplication.
    mov     r6, #10         // For radix-10 multiplication.
    
    // Skip any leading whitespace, comma, semicolon,
    // or colon characters in the string:
    
skipDelim:
    ldrb    r2, [r0], #1
    ldrb    r1, [r4, r2]    // Get classification
    cmp     r1, #delimCode  // Is it a delimiter?
    beq     skipDelim
    
    // First non-delimiter character must
    // be a valid decimal digit:
    
    blo     firstDigit        // Decimal digit
            
    // If the first digit is bad (including underscore or
	// zero-terminating byte), return -1:
    
badFirst:
    sub     r0, r0, #1          // Point back at bad char
    mov     r1, #-1
    cmp     r1, r1              // Sets carry flag
    rtn 
    
// Transfer here when the delimiter is the first
// character of a two-character sequence:

delimiter:
    sub     r0, r0, #2          // Point back at delimiter
    adds    r0, r0, #0          // Clears carry flag
    rtn
	
// If the non-decimal-digit character
// is bad, return -3:

illegal:
    mov     r1, #-3
    cmp     r1, r1              // Sets carry flag
    rtn
   
// Branch here if overflow occurs during
// the conversion:
 
overflow:
    mov     r1, #-2
    cmp     r1, r1              // Sets carry flag
    rtn
	
	.pool
	
// To avoid the very rare case where this code
// accesses data beyond an MMU page that is not
// present in memory, the following ensures
// that the source address is even:

firstDigit:
	tst		r0, #1				// Even if LO bit is 0.
	beq		evenPtr
	
	// If the current pointer value is odd, then
	// the first character (already read) was fetched
	// from an even address; back up and ignore the
	// previous processing of the first character in
	// order to align the pointer at an even address.
	
	sub		r0, r0, #1			// Redo first character.
	mov		r1, #0				// Reset accumulator to 0.

// At this point, the source pointer (R0) holds an even
// address, so we can safely process two characters at
// a time from memory. Also, half-word loads are always
// read from an even address, thus improving performance.

evenPtr:
	ldr		r4, =stateTable	
	ldrh	r2, [r0], #2			// Read two chars.
	ldr		r2, [r4, r2, lsl #2]	// Get address.
	bx		r2						// Branch to address.            

// State machine special cases where one (or both) of the two
// input characters are not decimal digits. Begin with
// the case where one character is a digit and the other
// character is an underscore:

	.macro	digitUS, digit
_\digit:
	cmp  r1, r3
	bhs  overflow
	mul  r1, r1, r6 //R6=10
	adds r1, r1, #\digit
	bcs	overflow
	ldrh r2, [r0], #2
	ldr  r2, [r4, r2, lsl #2]
	bx   r2
	.endm

// Special case for _0 (no add instrution needed):

	
	digitUS	0
	digitUS	1
	digitUS	2
	digitUS	3
	digitUS	4
	digitUS	5
	digitUS	6
	digitUS	7
	digitUS	8
	digitUS	9
	
	
// Handle the cases where the first character is
// a decimal digit and the second character is
// a delimiter (terminating the conversion).

_0Delimiter:
	cmp		r1, r3
	bhs		overflow
	mul		r1, r1, r6 //R6=10
	rtn

	.macro	digDelim digit
_\digit\()Delimiter:
	cmp		r1, r3
	bhs		overflow
	mul		r1, r1, r6
	adds	r1, r1, #\digit
	bcs		overflow
	rtn
	.endm

	digDelim 1
	digDelim 2
	digDelim 3
	digDelim 4
	digDelim 5
	digDelim 6
	digDelim 7
	digDelim 8
	digDelim 9
	
	.pool

// State machine for pairs of digits.

	.macro	digitPair, value
	.if		\value < 10
_0\value:
	.else
_\value:
	.endif
	cmp  r1, r3
	bhs  overflow
	mul  r1, r1, r5 //R5=100
	add  r1, r1, #\value
	bcs	 overflow
	ldrh r2, [r0], #2
	ldr  r2, [r4, r2, lsl #2]
	bx   r2
	.endm
	
	digitPair 0
	digitPair 1
	digitPair 2
	digitPair 3
	digitPair 4
	digitPair 5
	digitPair 6
	digitPair 7
	digitPair 8
	digitPair 9
	digitPair 10
	digitPair 11
	digitPair 12
	digitPair 13
	digitPair 14
	digitPair 15
	digitPair 16
	digitPair 17
	digitPair 18
	digitPair 19
	digitPair 20
	digitPair 21
	digitPair 22
	digitPair 23
	digitPair 24
	digitPair 25
	digitPair 26
	digitPair 27
	digitPair 28
	digitPair 29
	digitPair 30
	digitPair 31
	digitPair 32
	digitPair 33
	digitPair 34
	digitPair 35
	digitPair 36
	digitPair 37
	digitPair 38
	digitPair 39
	digitPair 40
	digitPair 41
	digitPair 42
	digitPair 43
	digitPair 44
	digitPair 45
	digitPair 46
	digitPair 47
	digitPair 48
	digitPair 49
	digitPair 50
	digitPair 51
	digitPair 52
	digitPair 53
	digitPair 54
	digitPair 55
	digitPair 56
	digitPair 57
	digitPair 58
	digitPair 59
	digitPair 60
	digitPair 61
	digitPair 62
	digitPair 63
	digitPair 64
	digitPair 65
	digitPair 66
	digitPair 67
	digitPair 68
	digitPair 69
	digitPair 70
	digitPair 71
	digitPair 72
	digitPair 73
	digitPair 74
	digitPair 75
	digitPair 76
	digitPair 77
	digitPair 78
	digitPair 79
	digitPair 80
	digitPair 81
	digitPair 82
	digitPair 83
	digitPair 84
	digitPair 85
	digitPair 86
	digitPair 87
	digitPair 88
	digitPair 89
	digitPair 90
	digitPair 91
	digitPair 92
	digitPair 93
	digitPair 94
	digitPair 95
	digitPair 96
	digitPair 97
	digitPair 98
	digitPair 99


    endp    bufToU32

    // Classify each character:
    //
    //  0 to 9-     decimal character
    //  0x10-       underscore
    //  0x11-       delimiter character
    //  0xff-       illegal character
    
classTbl:
    .byte   illChar     // 0x00
    .byte   illChar     // 0x01
    .byte   illChar     // 0x02
    .byte   illChar     // 0x03
    .byte   illChar     // 0x04
    .byte   illChar     // 0x05
    .byte   illChar     // 0x06
    .byte   illChar     // 0x07
    .byte   illChar     // 0x08
    .byte   delimCode   // 0x09
    .byte   delimCode   // 0x0a
    .byte   illChar     // 0x0b
    .byte   illChar     // 0x0c
    .byte   delimCode   // 0x0d
    .byte   illChar     // 0x0e
    .byte   illChar     // 0x0f
    .byte   illChar     // 0x10
    .byte   illChar     // 0x11
    .byte   illChar     // 0x12
    .byte   illChar     // 0x13
    .byte   illChar     // 0x14
    .byte   illChar     // 0x15
    .byte   illChar     // 0x16
    .byte   illChar     // 0x17
    .byte   illChar     // 0x18
    .byte   illChar     // 0x19
    .byte   illChar     // 0x1a
    .byte   illChar     // 0x1b
    .byte   illChar     // 0x1c
    .byte   illChar     // 0x1d
    .byte   illChar     // 0x1e
    .byte   illChar     // 0x1f
    .byte   delimCode   // ' '
    .byte   illChar     // '!'
    .byte   illChar     // '"'
    .byte   illChar     // '#'
    .byte   illChar     // '$'
    .byte   illChar     // '%'
    .byte   illChar     // '&'
    .byte   illChar     // '''
    .byte   illChar     // '('
    .byte   illChar     // ')'
    .byte   illChar     // '*'
    .byte   illChar     // '+'
    .byte   delimCode   // ','
    .byte   illChar     // '-'
    .byte   illChar     // '.'
    .byte   illChar     // '/'
    .byte   0           // '0'
    .byte   1           // '1'
    .byte   2           // '2'
    .byte   3           // '3'
    .byte   4           // '4'
    .byte   5           // '5'
    .byte   6           // '6'
    .byte   7           // '7'
    .byte   8           // '8'
    .byte   9           // '9'
    .byte   delimCode   // ':'
    .byte   delimCode   // ';'
    .byte   illChar     // '<'
    .byte   illChar     // '='
    .byte   illChar     // '>'
    .byte   illChar     // '?'
    .byte   illChar     // '@'
    .byte   illChar     // 'A'
    .byte   illChar     // 'B'
    .byte   illChar     // 'C'
    .byte   illChar     // 'D'
    .byte   illChar     // 'E'
    .byte   illChar     // 'F'
    .byte   illChar     // 'G'
    .byte   illChar     // 'H'
    .byte   illChar     // 'I'
    .byte   illChar     // 'J'
    .byte   illChar     // 'K'
    .byte   illChar     // 'L'
    .byte   illChar     // 'M'
    .byte   illChar     // 'N'
    .byte   illChar     // 'O'
    .byte   illChar     // 'P'
    .byte   illChar     // 'Q'
    .byte   illChar     // 'R'
    .byte   illChar     // 'S'
    .byte   illChar     // 'T'
    .byte   illChar     // 'U'
    .byte   illChar     // 'V'
    .byte   illChar     // 'W'
    .byte   illChar     // 'X'
    .byte   illChar     // 'Y'
    .byte   illChar     // 'Z'
    .byte   illChar     // '['
    .byte   illChar     // '\'
    .byte   illChar     // ']'
    .byte   illChar     // '^'
    .byte   underscore  // '_'
    .byte   illChar     // '`'
    .byte   illChar     // 'a'
    .byte   illChar     // 'b'
    .byte   illChar     // 'c'
    .byte   illChar     // 'd'
    .byte   illChar     // 'e'
    .byte   illChar     // 'f'
    .byte   illChar     // 'g'
    .byte   illChar     // 'h'
    .byte   illChar     // 'i'
    .byte   illChar     // 'j'
    .byte   illChar     // 'k'
    .byte   illChar     // 'l'
    .byte   illChar     // 'm'
    .byte   illChar     // 'n'
    .byte   illChar     // 'o'
    .byte   illChar     // 'p'
    .byte   illChar     // 'q'
    .byte   illChar     // 'r'
    .byte   illChar     // 's'
    .byte   illChar     // 't'
    .byte   illChar     // 'u'
    .byte   illChar     // 'v'
    .byte   illChar     // 'w'
    .byte   illChar     // 'x'
    .byte   illChar     // 'y'
    .byte   illChar     // 'z'
    .byte   illChar     // '{'
    .byte   illChar     // '|'
    .byte   illChar     // '}'
    .byte   illChar     // '~'
    .byte   illChar     // del
    .space  128, illChar    // 0x80 to 0xFF
	
stateTable:
	#include "stateTable.inc"


/////////////////////////////////////////////////////////////////////
//
// strToH32
//
// Converts the string pointed at by R0 to a numeric value
// and returns that numeric value in R1. This function
// preserves R0 (leaves it pointing at the beginning of the
// string buffer).

    proc    strToU32, thumb, public
    push    {r0, lr}
    bl      bufToU32
    pop     {r0, pc}
    endp    strToU32
            
/////////////////////////////////////////////////////////////////////
//              
// Here's the assembly language main program:
        
            proc    asmMain, public
            push    {r0-r8, lr}

// Timing code:

    ldr     r8, =iterations
timingLp:

    ldr     r7, =strPtrs
    mov     r6, #evCnt
innerLp:     
    ldr     r0, [r7], #4
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1:
    
    bl      bufToU32
            
    // Repeat for each sample input:
      
    subs    r6, #1
    bne     innerLp            

    // Repeat a large number of times in order
    // to time the code:
    
    subs    r8, r8, #1
    bne     timingLp


// Output code for testing:

    ldr     r7, =strPtrs
    mov     r6, #evCnt
convLp:     
    ldr     r0, [r7], #4
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1 and output
    // that value:
    
    bl      strToU32
    bcc     goodConv
    
    bl      print
    wastr   "Conversion error processing \"%s\", code = %ld\n"
    b       nextConv

goodConv:
    bl      print
    wastr   "Converted '%s' to %lu\n"
          
nextConv:
    subs    r6, #1
    bne     convLp
    
// Check for expected errors:
//
// Check empty string on input:

    ldr     r0, =eve
    bl      strToU32
    bcs     goodErr1
    bl      print
    wastr   "Expected error for empty string input\n"
    b       tryEvf
    
goodErr1:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"

// Check "*1234" on input:
    
tryEvf:     
    ldr     r0, =evf
    bl      strToU32
    bcs     goodErr2
    bl      print
    wastr   "Expected illegal 1st char for \"%s\"\n"
    b       tryEvg
    
goodErr2:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"

// Check "12*34" on input:
    
tryEvg:     
    ldr     r0, =evg
    bl      strToU32
    bcs     goodErr3
    bl      print
    wastr   "Expected illegal 1st char for \"%s\"n"
    b       tryEvh
    
goodErr3:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"

// Check for overflow:
    
tryEvh:     
    ldr     r0, =evh
    bl      strToU32
    bcs     goodErr4
    bl      print
    wastr   "Expected overflow for \"%s\" (%lx)\n"
    b       allDone
    
goodErr4:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"
                
allDone:             
    pop     {r0-r8, pc} 
    endp    asmMain


